@@ -19,6 +19,20 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/**
+ * SECTION:gsttagvorbis
+ * @short_description: tag mappings and support functions for plugins
+ *                     dealing with vorbiscomments
+ * @see_also: #GstTagList
+ *
+ * <refsect2>
+ * <para>
+ * Contains various utility functions for plugins to parse or create
+ * vorbiscomments and map them to and from #GstTagList<!-- -->s.
+ * </para>
+ * </refsect2>
+ */
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -27,7 +41,10 @@
 #include <stdlib.h>
 #include <string.h>
 
-static GstTagEntryMatch tag_matches[] = {
+/*
+ * see http://xiph.org/ogg/vorbis/doc/v-comment.html
+ */
+static const GstTagEntryMatch tag_matches[] = {
   {GST_TAG_TITLE, "TITLE"},
   {GST_TAG_VERSION, "VERSION"},
   {GST_TAG_ALBUM, "ALBUM"},
@@ -37,8 +54,11 @@ static GstTagEntryMatch tag_matches[] = {
   {GST_TAG_ALBUM_VOLUME_COUNT, "DISCTOTAL"},
   {GST_TAG_ARTIST, "ARTIST"},
   {GST_TAG_PERFORMER, "PERFORMER"},
+  {GST_TAG_COMPOSER, "COMPOSER"},
   {GST_TAG_COPYRIGHT, "COPYRIGHT"},
   {GST_TAG_LICENSE, "LICENSE"},
+  {GST_TAG_LICENSE_URI, "LICENSE"},
+  {GST_TAG_LOCATION, "LOCATION"},
   {GST_TAG_ORGANIZATION, "ORGANIZATION"},
   {GST_TAG_DESCRIPTION, "DESCRIPTION"},
   {GST_TAG_GENRE, "GENRE"},
@@ -50,6 +70,26 @@ static GstTagEntryMatch tag_matches[] = {
   {GST_TAG_TRACK_PEAK, "REPLAYGAIN_TRACK_PEAK"},
   {GST_TAG_ALBUM_GAIN, "REPLAYGAIN_ALBUM_GAIN"},
   {GST_TAG_ALBUM_PEAK, "REPLAYGAIN_ALBUM_PEAK"},
+  {GST_TAG_REFERENCE_LEVEL, "REPLAYGAIN_REFERENCE_LOUDNESS"},
+  {GST_TAG_MUSICBRAINZ_TRACKID, "MUSICBRAINZ_TRACKID"},
+  {GST_TAG_MUSICBRAINZ_ARTISTID, "MUSICBRAINZ_ARTISTID"},
+  {GST_TAG_MUSICBRAINZ_ALBUMID, "MUSICBRAINZ_ALBUMID"},
+  {GST_TAG_MUSICBRAINZ_ALBUMARTISTID, "MUSICBRAINZ_ALBUMARTISTID"},
+  {GST_TAG_MUSICBRAINZ_TRMID, "MUSICBRAINZ_TRMID"},
+  {GST_TAG_ARTIST_SORTNAME, "ARTISTSORT"},
+  {GST_TAG_ARTIST_SORTNAME, "ARTISTSORTORDER"},
+  {GST_TAG_ARTIST_SORTNAME, "MUSICBRAINZ_SORTNAME"},
+  {GST_TAG_ALBUM_SORTNAME, "ALBUMSORT"},
+  {GST_TAG_ALBUM_SORTNAME, "ALBUMSORTORDER"},
+  {GST_TAG_TITLE_SORTNAME, "TITLESORT"},
+  {GST_TAG_TITLE_SORTNAME, "TITLESORTORDER"},
+  {GST_TAG_LANGUAGE_CODE, "LANGUAGE"},
+  {GST_TAG_CDDA_MUSICBRAINZ_DISCID, "MUSICBRAINZ_DISCID"},
+  {GST_TAG_CDDA_CDDB_DISCID, "DISCID"},
+  /* some incidence that this makes sense:
+   * http://mail.kde.org/pipermail/amarok/2006-May/000090.html
+   */
+  {GST_TAG_BEATS_PER_MINUTE, "BPM"},
   {NULL, NULL}
 };
 
@@ -57,7 +97,7 @@ static GstTagEntryMatch tag_matches[] = {
  * gst_tag_from_vorbis_tag:
  * @vorbis_tag: vorbiscomment tag to convert to GStreamer tag
  *
- * Looks up the GStreamer tag for a vorbiscommment tag.
+ * Looks up the GStreamer tag for a vorbiscomment tag.
  *
  * Returns: The corresponding GStreamer tag or NULL if none exists.
  */
@@ -69,6 +109,8 @@ gst_tag_from_vorbis_tag (const gchar * vorbis_tag)
 
   g_return_val_if_fail (vorbis_tag != NULL, NULL);
 
+  gst_tag_register_musicbrainz_tags ();
+
   real_vorbis_tag = g_ascii_strup (vorbis_tag, -1);
   while (tag_matches[i].gstreamer_tag != NULL) {
     if (strcmp (real_vorbis_tag, tag_matches[i].original_tag) == 0) {
@@ -84,9 +126,9 @@ gst_tag_from_vorbis_tag (const gchar * vorbis_tag)
  * gst_tag_to_vorbis_tag:
  * @gst_tag: GStreamer tag to convert to vorbiscomment tag
  *
- * Looks up the vorbiscommment tag for a GStreamer tag.
+ * Looks up the vorbiscomment tag for a GStreamer tag.
  *
- * Returns: The corresponding vorbiscommment tag or NULL if none exists.
+ * Returns: The corresponding vorbiscomment tag or NULL if none exists.
  */
 G_CONST_RETURN gchar *
 gst_tag_to_vorbis_tag (const gchar * gst_tag)
@@ -95,6 +137,8 @@ gst_tag_to_vorbis_tag (const gchar * gst_tag)
 
   g_return_val_if_fail (gst_tag != NULL, NULL);
 
+  gst_tag_register_musicbrainz_tags ();
+
   while (tag_matches[i].gstreamer_tag != NULL) {
     if (strcmp (gst_tag, tag_matches[i].gstreamer_tag) == 0) {
       return tag_matches[i].original_tag;
@@ -105,14 +149,43 @@ gst_tag_to_vorbis_tag (const gchar * gst_tag)
 }
 
 
+/**
+ * gst_vorbis_tag_add:
+ * @list: a #GstTagList
+ * @tag: a vorbiscomment tag string (key in key=value), must be valid UTF-8
+ * @value: a vorbiscomment value string (value in key=value), must be valid UTF-8
+ *
+ * Convenience function using gst_tag_from_vorbis_tag(), parsing
+ * a vorbis comment string into the right type and adding it to the
+ * given taglist @list.
+ *
+ * Unknown vorbiscomment tags will be added to the tag list in form
+ * of a #GST_TAG_EXTENDED_COMMENT (since 0.10.10 at least).
+ */
 void
 gst_vorbis_tag_add (GstTagList * list, const gchar * tag, const gchar * value)
 {
-  const gchar *gst_tag = gst_tag_from_vorbis_tag (tag);
+  const gchar *gst_tag;
   GType tag_type;
 
-  if (gst_tag == NULL)
+  g_return_if_fail (list != NULL);
+  g_return_if_fail (tag != NULL);
+  g_return_if_fail (value != NULL);
+
+  g_return_if_fail (g_utf8_validate (tag, -1, NULL));
+  g_return_if_fail (g_utf8_validate (value, -1, NULL));
+  g_return_if_fail (strchr (tag, '=') == NULL);
+
+  gst_tag = gst_tag_from_vorbis_tag (tag);
+  if (gst_tag == NULL) {
+    gchar *ext_comment;
+
+    ext_comment = g_strdup_printf ("%s=%s", tag, value);
+    gst_tag_list_add (list, GST_TAG_MERGE_APPEND, GST_TAG_EXTENDED_COMMENT,
+        ext_comment, NULL);
+    g_free (ext_comment);
     return;
+  }
 
   tag_type = gst_tag_get_type (gst_tag);
   switch (tag_type) {
@@ -146,12 +219,29 @@ gst_vorbis_tag_add (GstTagList * list, const gchar * tag, const gchar * value)
       break;
     }
     case G_TYPE_STRING:{
-      gchar *valid;
+      gchar *valid = NULL;
 
-      if (!g_utf8_validate (value, -1, (const gchar **) &valid)) {
-        valid = g_strndup (value, valid - value);
-        GST_DEBUG ("Invalid vorbis comment tag, truncated it to %s\n", valid);
-      } else {
+      /* specialcase for language code */
+      if (strcmp (tag, "LANGUAGE") == 0) {
+        const gchar *s = strchr (value, '[');
+
+        /* Accept both ISO-639-1 and ISO-639-2 codes */
+        if (s && strchr (s, ']') == s + 4) {
+          valid = g_strndup (s + 1, 3);
+        } else if (s && strchr (s, ']') == s + 3) {
+          valid = g_strndup (s + 1, 2);
+        } else if (strlen (value) != 2 && strlen (value) != 3) {
+          GST_WARNING ("doesn't contain an ISO-639 language code: %s", value);
+        }
+      } else if (strcmp (tag, "LICENSE") == 0) {
+        /* license tags in vorbis comments must contain an URI representing
+         * the license and nothing more, at least according to:
+         * http://wiki.xiph.org/index.php/LICENSE_and_COPYRIGHT_tags_on_Vorbis_Comments */
+        if (value && gst_uri_is_valid (value))
+          gst_tag = GST_TAG_LICENSE_URI;
+      }
+
+      if (!valid) {
         valid = g_strdup (value);
       }
       gst_tag_list_add (list, GST_TAG_MERGE_APPEND, gst_tag, valid, NULL);
@@ -159,8 +249,13 @@ gst_vorbis_tag_add (GstTagList * list, const gchar * tag, const gchar * value)
       break;
     }
     case G_TYPE_DOUBLE:{
-      gst_tag_list_add (list, GST_TAG_MERGE_APPEND, gst_tag, g_strtod (value,
-              NULL), NULL);
+      gchar *c;
+
+      c = g_strdup (value);
+      g_strdelimit (c, ",", '.');
+      gst_tag_list_add (list, GST_TAG_MERGE_APPEND, gst_tag,
+          g_strtod (c, NULL), NULL);
+      g_free (c);
       break;
     }
     default:{
@@ -177,7 +272,18 @@ gst_vorbis_tag_add (GstTagList * list, const gchar * tag, const gchar * value)
             d = strtoul (check, &check, 10);
           }
         }
-        if (*check == '\0' && y != 0 && g_date_valid_dmy (d, m, y)) {
+
+        /* accept dates like 2007-00-00 and 2007-05-00 */
+        if (y != 0) {
+          if (m == 0 && d == 0)
+            m = d = 1;
+          else if (m != 0 && d == 0)
+            d = 1;
+        }
+
+        /* date might be followed by a time */
+        if ((*check == '\0' || g_ascii_isspace (*check)) && y != 0 &&
+            g_date_valid_dmy (d, m, y)) {
           GDate *date;
 
           date = g_date_new_dmy (d, m, y);
@@ -200,13 +306,13 @@ gst_vorbis_tag_add (GstTagList * list, const gchar * tag, const gchar * value)
  * @buffer: buffer to convert
  * @id_data: identification data at start of stream
  * @id_data_length: length of identification data
- * @ vendor_string: pointer to a string that should take the vendor string of this
- *                  vorbis comment or NULL if you don't need it.
+ * @vendor_string: pointer to a string that should take the vendor string
+ *                 of this vorbis comment or NULL if you don't need it.
  *
- * Creates a new tag list that contains the information parsed out of a 
+ * Creates a new tag list that contains the information parsed out of a
  * vorbiscomment packet.
  *
- * Returns: A new #GstTagList with all tags that could be extracted from the 
+ * Returns: A new #GstTagList with all tags that could be extracted from the
  *          given vorbiscomment buffer or NULL on error.
  */
 GstTagList *
@@ -231,16 +337,15 @@ gst_tag_list_from_vorbiscomment_buffer (const GstBuffer * buffer,
   GstTagList *list;
 
   g_return_val_if_fail (GST_IS_BUFFER (buffer), NULL);
-  g_return_val_if_fail (id_data != NULL, NULL);
-  g_return_val_if_fail (id_data_length > 0, NULL);
+  g_return_val_if_fail (id_data != NULL || id_data_length == 0, NULL);
 
   data = GST_BUFFER_DATA (buffer);
   size = GST_BUFFER_SIZE (buffer);
   list = gst_tag_list_new ();
 
-  if (size < 11)
+  if (size < 11 || size <= id_data_length + 4)
     goto error;
-  if (memcmp (data, id_data, id_data_length) != 0)
+  if (id_data_length > 0 && memcmp (data, id_data, id_data_length) != 0)
     goto error;
   ADVANCE (id_data_length);
   if (vendor_string)
@@ -282,16 +387,35 @@ typedef struct
 }
 MyForEach;
 
+/**
+ * gst_tag_to_vorbis_comments:
+ * @list: a #GstTagList
+ * @tag: a GStreamer tag identifier, such as #GST_TAG_ARTIST
+ *
+ * Creates a new tag list that contains the information parsed out of a
+ * vorbiscomment packet.
+ *
+ * Returns: A #GList of newly-allowcated key=value strings. Free with
+ *          g_list_foreach (list, (GFunc) g_free, NULL) plus g_list_free (list)
+ */
 GList *
 gst_tag_to_vorbis_comments (const GstTagList * list, const gchar * tag)
 {
+  const gchar *vorbis_tag = NULL;
   GList *l = NULL;
   guint i;
-  const gchar *vorbis_tag = gst_tag_to_vorbis_tag (tag);
 
-  if (!vorbis_tag)
-    return NULL;
+  g_return_val_if_fail (list != NULL, NULL);
+  g_return_val_if_fail (tag != NULL, NULL);
 
+  if (strcmp (tag, GST_TAG_EXTENDED_COMMENT) != 0) {
+    vorbis_tag = gst_tag_to_vorbis_tag (tag);
+    if (!vorbis_tag)
+      return NULL;
+  }
+
+  /* FIXME: for tags that can map to multiple vorbis comment keys, add all
+   * of the possible keys */
   for (i = 0; i < gst_tag_list_get_tag_size (list, tag); i++) {
     GType tag_type = gst_tag_get_type (tag);
     gchar *result = NULL;
@@ -306,21 +430,36 @@ gst_tag_to_vorbis_comments (const GstTagList * list, const gchar * tag)
         break;
       }
       case G_TYPE_STRING:{
-        gchar *str;
+        gchar *str = NULL;
 
         if (!gst_tag_list_get_string_index (list, tag, i, &str))
           g_return_val_if_reached (NULL);
-        result = g_strdup_printf ("%s=%s", vorbis_tag, str);
+
+        /* special case: GST_TAG_EXTENDED_COMMENT */
+        if (vorbis_tag == NULL) {
+          gchar *key = NULL, *val = NULL;
+
+          if (gst_tag_parse_extended_comment (str, &key, NULL, &val, TRUE)) {
+            result = g_strdup_printf ("%s=%s", key, val);
+            g_free (key);
+            g_free (val);
+          } else {
+            GST_WARNING ("Not a valid extended comment string: %s", str);
+          }
+        } else {
+          result = g_strdup_printf ("%s=%s", vorbis_tag, str);
+        }
         g_free (str);
         break;
       }
       case G_TYPE_DOUBLE:{
         gdouble value;
+        gchar buf[G_ASCII_DTOSTR_BUF_SIZE];
 
         if (!gst_tag_list_get_double_index (list, tag, i, &value))
           g_return_val_if_reached (NULL);
-        /* FIXME: what about locale-specific floating point separators? */
-        result = g_strdup_printf ("%s=%f", vorbis_tag, value);
+        g_ascii_formatd (buf, G_ASCII_DTOSTR_BUF_SIZE, "%f", value);
+        result = g_strconcat (vorbis_tag, "=", buf, NULL);
         break;
       }
       default:{
@@ -364,19 +503,21 @@ write_one_tag (const GstTagList * list, const gchar * tag, gpointer user_data)
     data->data_count += strlen (result);
     data->entries = g_list_prepend (data->entries, result);
   }
+
+  g_list_free (comments);
 }
 
 /**
  * gst_tag_list_to_vorbiscomment_buffer:
- * @buffer: tag list to convert
+ * @list: tag list to convert
  * @id_data: identification data at start of stream
- * @id_data_length: length of identification data
- * @ vendor_string: string that describes the vendor string or NULL
+ * @id_data_length: length of identification data, may be 0 if @id_data is NULL
+ * @vendor_string: string that describes the vendor string or NULL
  *
  * Creates a new vorbiscomment buffer from a tag list.
  *
- * Returns: A new #GstBuffer containing a vorbiscomment buffer with all tags that 
- *          could be converted from the given tag list.
+ * Returns: A new #GstBuffer containing a vorbiscomment buffer with all tags
+ *          that could be converted from the given tag list.
  */
 GstBuffer *
 gst_tag_list_to_vorbiscomment_buffer (const GstTagList * list,
@@ -392,8 +533,7 @@ gst_tag_list_to_vorbiscomment_buffer (const GstTagList * list,
   int required_size;
 
   g_return_val_if_fail (GST_IS_TAG_LIST (list), NULL);
-  g_return_val_if_fail (id_data != NULL, NULL);
-  g_return_val_if_fail (id_data_length > 0, NULL);
+  g_return_val_if_fail (id_data != NULL || id_data_length == 0, NULL);
 
   if (vendor_string == NULL)
     vendor_string = "GStreamer encoded vorbiscomment";
@@ -403,14 +543,16 @@ gst_tag_list_to_vorbiscomment_buffer (const GstTagList * list,
   required_size += 4 * my_data.count + my_data.data_count;
   buffer = gst_buffer_new_and_alloc (required_size);
   data = GST_BUFFER_DATA (buffer);
-  memcpy (data, id_data, id_data_length);
-  data += id_data_length;
-  *((guint32 *) data) = GUINT32_TO_LE (vendor_len);
+  if (id_data_length > 0) {
+    memcpy (data, id_data, id_data_length);
+    data += id_data_length;
+  }
+  GST_WRITE_UINT32_LE (data, vendor_len);
   data += 4;
   memcpy (data, vendor_string, vendor_len);
   data += vendor_len;
   l = my_data.entries = g_list_reverse (my_data.entries);
-  *((guint32 *) data) = GUINT32_TO_LE (my_data.count);
+  GST_WRITE_UINT32_LE (data, my_data.count);
   data += 4;
   for (i = 0; i < my_data.count; i++) {
     guint size;
@@ -420,7 +562,7 @@ gst_tag_list_to_vorbiscomment_buffer (const GstTagList * list,
     cur = l->data;
     l = g_list_next (l);
     size = strlen (cur);
-    *((guint32 *) data) = GUINT32_TO_LE (size);
+    GST_WRITE_UINT32_LE (data, size);
     data += 4;
     memcpy (data, cur, size);
     data += size;
